module Transpiler
using CUDA
using ..ExpressionProcessing

# culoadtest(N, rand(["add.f32", "sub.f32", "mul.f32", "div.approx.f32"]))
function culoadtest(N::Int32, op = "add.f32")

	vadd_code = ".version 7.1
	
	.target sm_52
	.address_size 64
	
		// .globl       VecAdd_kernel
	
	.visible .entry VecAdd_kernel(
		.param .u64 VecAdd_kernel_param_0,
		.param .u64 VecAdd_kernel_param_1,
		.param .u64 VecAdd_kernel_param_2,
		.param .u32 VecAdd_kernel_param_3
	)
	
	{
		.reg .pred      %p<2>;
		.reg .f32       %f<4>;
		.reg .b32       %r<6>;
		.reg .b64       %rd<11>;

		ld.param.u64    %rd1, [VecAdd_kernel_param_0];
		ld.param.u64    %rd2, [VecAdd_kernel_param_1];
		ld.param.u64    %rd3, [VecAdd_kernel_param_2];
		ld.param.u32    %r2, [VecAdd_kernel_param_3];

		mov.u32         %r3, %ntid.x;
		mov.u32         %r4, %ctaid.x;
		mov.u32         %r5, %tid.x;

		mad.lo.s32      %r1, %r3, %r4, %r5;

		setp.ge.s32     %p1, %r1, %r2;

		@%p1 bra        \$L__BB0_2;

		cvta.to.global.u64      %rd4, %rd1;

		mul.wide.s32    %rd5, %r1, 4;
		add.s64         %rd6, %rd4, %rd5;
		cvta.to.global.u64      %rd7, %rd2;
		add.s64         %rd8, %rd7, %rd5;

		ld.global.f32   %f1, [%rd8];
		ld.global.f32   %f2, [%rd6];" *
		op *
		"               %f3, %f2, %f1;
		cvta.to.global.u64      %rd9, %rd3;
		add.s64         %rd10, %rd9, %rd5;
		st.global.f32   [%rd10], %f3;
	
	\$L__BB0_2:
		ret;
	}"
	
		linker = CuLink()
		add_data!(linker, "VecAdd_kernel", vadd_code)
	
		image = complete(linker)
	
		mod = CuModule(image)
		func = CuFunction(mod, "VecAdd_kernel")
	
		d_a = CUDA.fill(1.0f0, N)
		d_b = CUDA.fill(2.0f0, N)
		d_c = CUDA.fill(0.0f0, N)
	
		# Grid/Block configuration
		threadsPerBlock = 256;
		blocksPerGrid  = (N + threadsPerBlock - 1) ÷ threadsPerBlock;
	
		@time CUDA.@sync cudacall(func, Tuple{CuPtr{Cfloat},CuPtr{Cfloat},CuPtr{Cfloat},Cint}, d_a, d_b, d_c, N; threads=threadsPerBlock, blocks=blocksPerGrid)
end

const exitJumpLocationMarker = "\$L__BB0_2"
function transpile(expression::ExpressionProcessing.PostfixType)
	ptxBuffer = IOBuffer()

	println(ptxBuffer, get_cuda_header())
	println(ptxBuffer, get_kernel_signature("ExpressionProcessing", [Int64, Float64]))
	println(ptxBuffer, "{")

	# Register definition
	println(ptxBuffer, get_register_definitions(1, 5)) # TODO: Actually calculate the number of needed registers and extend to more register kinds
	# Parameter loading
	println(ptxBuffer, get_guard_clause())

	# Code goes here

	# exit jump location
	print(ptxBuffer, exitJumpLocationMarker); println(ptxBuffer, ": ret;")
	println(ptxBuffer, "}")
	println(String(take!(ptxBuffer)))
end

# TODO: Make version, target and address_size configurable
function get_cuda_header()::String
	return "
.version 7.1
.target sm_52
.address_size 64
"
end

function get_kernel_signature(kernelName::String, parameters::Vector{DataType})::String
	signatureBuffer = IOBuffer()
	print(signatureBuffer, ".visible .entry ")
	print(signatureBuffer, kernelName)
	println(signatureBuffer, "(")

	
	for i in eachindex(parameters)
		type = type_to_cuda_type(parameters[i])
		print(signatureBuffer, 
			  "  .param ", type, " ", kernelName, "_param_", i)
		if i != lastindex(parameters)
			println(signatureBuffer, ",")
		end
	end

	print(signatureBuffer, ")")
	return String(take!(signatureBuffer))
end

"
Constructs the PTX code used for handling the case where too many threads are started.

Assumes the following:
 - There are the unused ```32 bit``` registers ```r1, r2, r3, r4 (index of the variable set)```
 - There is an unused ```predicate``` register ```p1```
 - The ```32 bit``` register ```r5``` contains the number of variable sets
"
function get_guard_clause()::String
	guardBuffer = IOBuffer()

	println(guardBuffer, "mov.u32    %r1, %ntid.x;") # nr of thread ids
	println(guardBuffer, "mov.u32    %r2, %ctaid.x;") # nr of threads per cta
	println(guardBuffer, "mov.u32    %r3, %tid.x;") # id of the current thread

	println(guardBuffer, "mad.lo.s32    %r4, %r1, %r2, %r3;") # the current index (basically index of variable set)
	println(guardBuffer, "setp.ge.s32    %p1, %r4, %r5;") # guard clause (p1 = r4 > r5 -> index > nrOfVariableSets)

	# branch to end if p1 is true
	print(guardBuffer, "@%p1 bra    ")
	print(guardBuffer, exitJumpLocationMarker)
	println(guardBuffer, ";")

	return String(take!(guardBuffer))
end

function get_register_definitions(nrPred::Int, nr32Bit::Int):::String
	registersBuffer = IOBuffer()

	if nrPred > 0
		print(registersBuffer, ".reg .pred")
		print(registersBuffer, "    %p<")
		print(registersBuffer, nrPred + 1)
		println(registersBuffer, ">;")
	end
	if nr32Bit > 0
		print(registersBuffer, ".reg .b32")
		print(registersBuffer, "    %r<")
		print(registersBuffer, nr32Bit + 1)
		println(registersBuffer, ">;")
	end

	return String(take!(registersBuffer))
end

function type_to_cuda_type(type::DataType)::String
	if type == Int64
		return ".s64"
	elseif type == Float64
		return ".f64"
	else
		return ".b64"
	end
end


end

